@@ -119,8 +119,20 @@ class LinuxSCSI(executor.Executor):
 
         return dev_info
 
-    def get_sysfs_wwn(self, device_names):
+    def get_sysfs_wwn(self, device_names, mpath=None):
         """Return the wwid from sysfs in any of devices in udev format."""
+        # If we have a multipath DM we know that it has found the WWN
+        if mpath:
+            # We have the WWN in /uuid even with friendly names, unline /name
+            try:
+                with open('/sys/block/%s/dm/uuid' % mpath) as f:
+                    # Contents are matph-WWN, so get the part we want
+                    wwid = f.read().strip()[6:]
+                    if wwid:  # Check should not be needed, but just in case
+                        return wwid
+            except Exception as exc:
+                LOG.warning('Failed to read the DM uuid: %s', exc)
+
         wwid = self.get_sysfs_wwid(device_names)
         glob_str = '/dev/disk/by-id/scsi-'
         wwn_paths = glob.glob(glob_str + '*')
@@ -128,16 +140,34 @@ class LinuxSCSI(executor.Executor):
         if wwid and glob_str + wwid in wwn_paths:
             return wwid
 
-        # If we have multiple designators follow the symlinks
+        # If we have multiple designators use symlinks to find out the wwn
+        device_names = set(device_names)
         for wwn_path in wwn_paths:
             try:
                 if os.path.islink(wwn_path) and os.stat(wwn_path):
                     path = os.path.realpath(wwn_path)
-                    if path.startswith('/dev/') and path[5:] in device_names:
-                        return wwn_path[len(glob_str):]
+                    if path.startswith('/dev/'):
+                        name = path[5:]
+                        # Symlink may point to the multipath dm if the attach
+                        # was too fast or we took long to check it. Check
+                        # devices belonging to the multipath DM.
+                        if name.startswith('dm-'):
+                            # Get the devices that belong to the DM
+                            slaves_path = '/sys/class/block/%s/slaves' % name
+                            dm_devs = os.listdir(slaves_path)
+                            # This is the right wwn_path if the devices we have
+                            # attached belong to the dm we followed
+                            if device_names.intersection(dm_devs):
+                                break
+
+                        # This is the right wwn_path if  devices we have
+                        elif name in device_names:
+                            break
             except OSError:
                 continue
-        return ''
+        else:
+            return ''
+        return wwn_path[len(glob_str):]
 
     def get_sysfs_wwid(self, device_names):
         """Return the wwid from sysfs in any of devices in udev format."""
