@@ -1111,6 +1111,27 @@ Setting up iSCSI targets: unused
             (('ip1:port1', 'tgt1', 4),),
             force=True, ignore_errors=True)
 
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn', return_value='')
+    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
+    @mock.patch.object(iscsi.ISCSIConnector, '_cleanup_connection')
+    @mock.patch('time.sleep')
+    def test_connect_single_volume_no_wwn(self, sleep_mock, cleanup_mock,
+                                          connect_mock, get_wwn_mock):
+        def my_connect(rescans, props, data):
+            data['found_devices'].append('sdz')
+
+        connect_mock.side_effect = my_connect
+
+        res = self.connector._connect_single_volume(self.CON_PROPS)
+
+        expected = {'type': 'block', 'scsi_wwn': '', 'path': '/dev/sdz'}
+        self.assertEqual(expected, res)
+        get_wwn_mock.assert_has_calls([mock.call(['sdz'])] * 10)
+        self.assertEqual(10, get_wwn_mock.call_count)
+        sleep_mock.assert_has_calls([mock.call(1)] * 10)
+        self.assertEqual(10, sleep_mock.call_count)
+        cleanup_mock.assert_not_called()
+
     @staticmethod
     def _get_connect_vol_data():
         return {'stop_connecting': False, 'num_logins': 0, 'failed_logins': 0,
@@ -1185,11 +1206,52 @@ Setting up iSCSI targets: unused
         result = list(get_wwn_mock.call_args[0][0])
         result.sort()
         self.assertEqual(['sda', 'sdb', 'sdc', 'sdd'], result)
+        # Check we pass the mpath
+        self.assertIsNone(get_wwn_mock.call_args[0][1])
         add_wwid_mock.assert_called_once_with('wwn')
         self.assertNotEqual(0, add_path_mock.call_count)
         self.assertGreaterEqual(find_dm_mock.call_count, 2)
         self.assertEqual(4, connect_mock.call_count)
 
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_sysfs_multipath_dm',
+                       side_effect=[None, 'dm-0'])
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn', return_value='')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_path')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_wwid')
+    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
+    @mock.patch('time.sleep')
+    def test_connect_multipath_volume_no_wwid(self, sleep_mock, connect_mock,
+                                              add_wwid_mock, add_path_mock,
+                                              get_wwn_mock, find_dm_mock):
+        # Even if we don't have the wwn we'll be able to find the multipath
+        def my_connect(rescans, props, data):
+            devs = {'tgt1': 'sda', 'tgt2': 'sdb', 'tgt3': 'sdc', 'tgt4': 'sdd'}
+            data['stopped_threads'] += 1
+            data['num_logins'] += 1
+            dev = devs[props['target_iqn']]
+            data['found_devices'].append(dev)
+            data['just_added_devices'].append(dev)
+
+        connect_mock.side_effect = my_connect
+
+        res = self.connector._connect_multipath_volume(self.CON_PROPS)
+
+        expected = {'type': 'block', 'scsi_wwn': '', 'multipath_id': '',
+                    'path': '/dev/dm-0'}
+        self.assertEqual(expected, res)
+
+        self.assertGreaterEqual(get_wwn_mock.call_count, 3)
+        result = list(get_wwn_mock.call_args[0][0])
+        result.sort()
+        self.assertEqual(['sda', 'sdb', 'sdc', 'sdd'], result)
+        # Initially mpath we pass is None, but on last call is the mpath
+        mpath_values = [c[1][1] for c in get_wwn_mock._mock_mock_calls]
+        self.assertEqual([None, None, 'dm-0'], mpath_values[0:3])
+        add_wwid_mock.assert_not_called()
+        add_path_mock.assert_not_called()
+        self.assertGreaterEqual(find_dm_mock.call_count, 2)
+        self.assertEqual(4, connect_mock.call_count)
+
     @mock.patch.object(linuxscsi.LinuxSCSI, 'find_sysfs_multipath_dm',
                        side_effect=[None, 'dm-0'])
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn',
@@ -1562,17 +1624,67 @@ Setting up iSCSI targets: unused
             mock.call('/dev/disk/by-id/dm-...'),
             mock.call('/dev/disk/by-id/scsi-wwn')])
 
+    @mock.patch('time.sleep')
     @mock.patch('os.path.realpath', return_value='/dev/sdz')
     @mock.patch('os.listdir', return_value=['dm-...', 'scsi-...'])
-    def test__get_device_link_not_found(self, listdir_mock, realpath_mock):
+    def test__get_device_link_not_found(self, listdir_mock, realpath_mock,
+                                        mock_time):
         self.assertRaises(exception.VolumeDeviceNotFound,
                           self.connector._get_device_link,
                           'wwn', '/dev/sda', None)
-        listdir_mock.assert_called_once_with('/dev/disk/by-id/')
-        realpath_mock.assert_has_calls([
-            mock.call('/dev/disk/by-id/scsi-wwn'),
-            mock.call('/dev/disk/by-id/dm-...'),
-            mock.call('/dev/disk/by-id/scsi-...')])
+        listdir_mock.assert_has_calls(3 * [mock.call('/dev/disk/by-id/')])
+        self.assertEqual(3, listdir_mock.call_count)
+        realpath_mock.assert_has_calls(
+            3 * [mock.call('/dev/disk/by-id/scsi-wwn'),
+                 mock.call('/dev/disk/by-id/dm-...'),
+                 mock.call('/dev/disk/by-id/scsi-...')])
+        self.assertEqual(9, realpath_mock.call_count)
+
+    @mock.patch('time.sleep')
+    @mock.patch('os.path.realpath')
+    @mock.patch('os.listdir', return_value=['dm-...', 'scsi-...'])
+    def test__get_device_link_symlink_found_after_retry(self, mock_listdir,
+                                                        mock_realpath,
+                                                        mock_time):
+        # Return the expected realpath on the third retry
+        mock_realpath.side_effect = [
+            None, None, None, None, None, None, '/dev/sda']
+
+        # Assert that VolumeDeviceNotFound isn't raised
+        self.connector._get_device_link('wwn', '/dev/sda', None)
+
+        # Assert that listdir and realpath have been called correctly
+        mock_listdir.assert_has_calls(2 * [mock.call('/dev/disk/by-id/')])
+        self.assertEqual(2, mock_listdir.call_count)
+        mock_realpath.assert_has_calls(
+            2 * [mock.call('/dev/disk/by-id/scsi-wwn'),
+                 mock.call('/dev/disk/by-id/dm-...'),
+                 mock.call('/dev/disk/by-id/scsi-...')]
+            + [mock.call('/dev/disk/by-id/scsi-wwn')])
+        self.assertEqual(7, mock_realpath.call_count)
+
+    @mock.patch('time.sleep')
+    @mock.patch('os.path.realpath')
+    @mock.patch('os.listdir', return_value=['dm-...', 'scsi-...'])
+    def test__get_device_link_symlink_found_after_retry_by_listdir(
+            self, mock_listdir, mock_realpath, mock_time):
+
+        # Return the expected realpath on the second retry while looping over
+        # the devices returned by listdir
+        mock_realpath.side_effect = [
+            None, None, None, None, None, '/dev/sda']
+
+        # Assert that VolumeDeviceNotFound isn't raised
+        self.connector._get_device_link('wwn', '/dev/sda', None)
+
+        # Assert that listdir and realpath have been called correctly
+        mock_listdir.assert_has_calls(2 * [mock.call('/dev/disk/by-id/')])
+        self.assertEqual(2, mock_listdir.call_count)
+        mock_realpath.assert_has_calls(
+            2 * [mock.call('/dev/disk/by-id/scsi-wwn'),
+                 mock.call('/dev/disk/by-id/dm-...'),
+                 mock.call('/dev/disk/by-id/scsi-...')])
+        self.assertEqual(6, mock_realpath.call_count)
 
     @mock.patch.object(iscsi.ISCSIConnector, '_run_iscsiadm_bare')
     def test_get_node_startup_values(self, run_iscsiadm_bare_mock):
@@ -1644,8 +1756,8 @@ Setting up iSCSI targets: unused
         iqn1 = 'iqn.2010-10.org.openstack:%s' % name1
         iqn2 = 'iqn.2010-10.org.openstack:%s' % name2
         iqn3 = 'iqn.2010-10.org.openstack:%s' % name3
-        connection_properties = self.iscsi_connection(vol, [location], [iqn1])
-        recover_connection = self.iscsi_connection(vol, [location], [iqn2])
+        connection_properties = self.iscsi_connection(vol, [location], iqn1)
+        recover_connection = self.iscsi_connection(vol, [location], iqn2)
 
         node_startup1 = "manual"
         node_startup2 = "automatic"
