@@ -761,7 +761,8 @@ loop0                                     0"""
         mock_find_mpath_path.return_value = ('/dev/mapper/dm-uuid-mpath-%s' %
                                              wwn)
 
-        ret_size = self.linuxscsi.extend_volume(['/dev/fake1', '/dev/fake2'])
+        ret_size = self.linuxscsi.extend_volume(['/dev/fake1', '/dev/fake2'],
+                                                use_multipath=True)
         self.assertEqual(2048, ret_size)
 
         # because we don't mock out the echo_scsi_command
@@ -795,7 +796,8 @@ loop0                                     0"""
 
         mock_mpath_resize_map.return_value = 'fail'
 
-        ret_size = self.linuxscsi.extend_volume(['/dev/fake1', '/dev/fake2'])
+        ret_size = self.linuxscsi.extend_volume(['/dev/fake1', '/dev/fake2'],
+                                                use_multipath=True)
         self.assertIsNone(ret_size)
 
         # because we don't mock out the echo_scsi_command
@@ -846,6 +848,16 @@ loop0                                     0"""
         mock_exec.assert_called_once_with(
             'multipathd', 'show', 'status', run_as_root=True, root_helper=None)
 
+    @mock.patch('six.moves.builtins.open')
+    def test_get_sysfs_wwn_mpath(self, open_mock):
+        wwn = '3600d0230000000000e13955cc3757800'
+        cm_open = open_mock.return_value.__enter__.return_value
+        cm_open.read.return_value = 'mpath-' + wwn
+
+        res = self.linuxscsi.get_sysfs_wwn(mock.sentinel.device_names, 'dm-1')
+        open_mock.assert_called_once_with('/sys/block/dm-1/dm/uuid')
+        self.assertEqual(wwn, res)
+
     @mock.patch('glob.glob')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwid')
     def test_get_sysfs_wwn_single_designator(self, get_wwid_mock, glob_mock):
@@ -857,25 +869,67 @@ loop0                                     0"""
         glob_mock.assert_called_once_with('/dev/disk/by-id/scsi-*')
         get_wwid_mock.assert_called_once_with(mock.sentinel.device_names)
 
+    @mock.patch('six.moves.builtins.open', side_effect=Exception)
+    @mock.patch('glob.glob')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwid')
+    def test_get_sysfs_wwn_mpath_exc(self, get_wwid_mock, glob_mock,
+                                     open_mock):
+        glob_mock.return_value = ['/dev/disk/by-id/scsi-wwid1',
+                                  '/dev/disk/by-id/scsi-wwid2']
+        get_wwid_mock.return_value = 'wwid1'
+        res = self.linuxscsi.get_sysfs_wwn(mock.sentinel.device_names, 'dm-1')
+        open_mock.assert_called_once_with('/sys/block/dm-1/dm/uuid')
+        self.assertEqual('wwid1', res)
+        glob_mock.assert_called_once_with('/dev/disk/by-id/scsi-*')
+        get_wwid_mock.assert_called_once_with(mock.sentinel.device_names)
+
+    @mock.patch('os.listdir', return_value=['sda', 'sdd'])
     @mock.patch('os.path.realpath', side_effect=('/other/path',
+                                                 '/dev/dm-5',
                                                  '/dev/sda', '/dev/sdb'))
-    @mock.patch('os.path.islink', side_effect=(False, True, True, True, True))
-    @mock.patch('os.stat', side_effect=(False, True, True, True))
+    @mock.patch('os.path.islink', side_effect=(False,) + (True,) * 5)
+    @mock.patch('os.stat', side_effect=(False,) + (True,) * 4)
     @mock.patch('glob.glob')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwid')
     def test_get_sysfs_wwn_multiple_designators(self, get_wwid_mock, glob_mock,
                                                 stat_mock, islink_mock,
-                                                realpath_mock):
+                                                realpath_mock, listdir_mock):
         glob_mock.return_value = ['/dev/disk/by-id/scsi-fail-link',
                                   '/dev/disk/by-id/scsi-fail-stat',
                                   '/dev/disk/by-id/scsi-non-dev',
+                                  '/dev/disk/by-id/scsi-another-dm',
                                   '/dev/disk/by-id/scsi-wwid1',
                                   '/dev/disk/by-id/scsi-wwid2']
+
         get_wwid_mock.return_value = 'pre-wwid'
         devices = ['sdb', 'sdc']
         res = self.linuxscsi.get_sysfs_wwn(devices)
         self.assertEqual('wwid2', res)
         glob_mock.assert_called_once_with('/dev/disk/by-id/scsi-*')
+        listdir_mock.assert_called_once_with('/sys/class/block/dm-5/slaves')
+        get_wwid_mock.assert_called_once_with(devices)
+
+    @mock.patch('os.listdir', side_effect=[['sda', 'sdb'], ['sdc', 'sdd']])
+    @mock.patch('os.path.realpath', side_effect=('/dev/sde',
+                                                 '/dev/dm-5',
+                                                 '/dev/dm-6'))
+    @mock.patch('os.path.islink', mock.Mock())
+    @mock.patch('os.stat', mock.Mock())
+    @mock.patch('glob.glob')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwid', return_value='')
+    def test_get_sysfs_wwn_dm_link(self, get_wwid_mock, glob_mock,
+                                   realpath_mock, listdir_mock):
+        glob_mock.return_value = ['/dev/disk/by-id/scsi-wwid1',
+                                  '/dev/disk/by-id/scsi-another-dm',
+                                  '/dev/disk/by-id/scsi-our-dm']
+
+        devices = ['sdc', 'sdd']
+        res = self.linuxscsi.get_sysfs_wwn(devices)
+        self.assertEqual('our-dm', res)
+        glob_mock.assert_called_once_with('/dev/disk/by-id/scsi-*')
+        listdir_mock.assert_has_calls(
+            [mock.call('/sys/class/block/dm-5/slaves'),
+             mock.call('/sys/class/block/dm-6/slaves')])
         get_wwid_mock.assert_called_once_with(devices)
 
     @mock.patch('os.path.realpath', side_effect=('/dev/sda', '/dev/sdb'))
@@ -894,6 +948,16 @@ loop0                                     0"""
         glob_mock.assert_called_once_with('/dev/disk/by-id/scsi-*')
         get_wwid_mock.assert_called_once_with(devices)
 
+    @mock.patch('glob.glob', return_value=[])
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwid')
+    def test_get_sysfs_wwn_no_links(self, get_wwid_mock, glob_mock):
+        get_wwid_mock.return_value = ''
+        devices = ['sdc']
+        res = self.linuxscsi.get_sysfs_wwn(devices)
+        self.assertEqual('', res)
+        glob_mock.assert_called_once_with('/dev/disk/by-id/scsi-*')
+        get_wwid_mock.assert_called_once_with(devices)
+
     @ddt.data({'wwn_type': 't10.', 'num_val': '1'},
               {'wwn_type': 'eui.', 'num_val': '2'},
               {'wwn_type': 'naa.', 'num_val': '3'})
@@ -920,18 +984,6 @@ loop0                                     0"""
         open_mock.assert_has_calls([mock.call('/sys/block/sda/device/wwid'),
                                     mock.call('/sys/block/sdb/device/wwid')])
 
-    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn')
-    def test_verify_sysfs_wwns_valid(self, get_wwn):
-        get_wwn.side_effect = ['123', '123', '123', '123']
-        self.linuxscsi.verify_sysfs_wwns(['sda', 'sdb', 'sdc', 'sdd'], '123')
-
-    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn')
-    def test_verify_sysfs_wwns_invalid(self, get_wwn):
-        get_wwn.side_effect = ['123', '456', '123', '123']
-        self.assertRaises(exception.VolumeDeviceValidationFailed,
-                          self.linuxscsi.verify_sysfs_wwns,
-                          ['sda', 'sdb', 'sdc', 'sdd'], '123')
-
     @mock.patch.object(linuxscsi.priv_rootwrap, 'unlink_root')
     @mock.patch('glob.glob')
     @mock.patch('os.path.realpath', side_effect=['/dev/sda', '/dev/sdb',
